import torch
import numpy as np
import random
from torchvision import datasets
from torchvision import transforms
from torch.utils.data import TensorDataset
from torch.utils.data.sampler import SubsetRandomSampler

class ReshapeTransform:
    def __init__(self, new_size):
        self.new_size = new_size

    def __call__(self, img):
        return torch.reshape(img, self.new_size)

class DataUtils:
    @staticmethod
    def ensure_tensor(a):
        if type(a) is torch.Tensor:
            return a
        elif type(a) is np.ndarray:
            return torch.from_numpy(a)
        else: # handle all other types convertible by numpy
            return torch.from_numpy(np.array(a))

    @staticmethod
    def channel_norm(ds, channel_dim=None):
        # collect tensors in list
        l = [DataUtils.ensure_tensor(data) for data, *_ in ds]
        # join back all tensors so the first dimension is count of tensors
        l = torch.stack(l, dim=0) #size: [N, X, Y, ...] or [N, C, X, Y, ...]

        if channel_dim is None:
            # add redundant first dim
            l = l.unsqueeze(0)

        else:
            # swap channel dimension to first
            l = torch.transpose(l, 0, channel_dim).contiguous() #size: [C, N, X, Y, ...]
        # collapse all except first dimension
        l = l.view(l.size(0), -1) #size: [C, N*X*Y]
        mean = torch.mean(l, dim=1) #size: [C]
        std = torch.std(l, dim=1) #size: [C]
        return (mean, std)

    @staticmethod
    def sample_by_class(ds, k, as_np=False, shuffle=False, no_test=False):
        class_counts = {}
        train_data = []
        train_label = []
        test_data = []
        test_label = []

        #TODO optimize this
        if shuffle:
            ds = [(d, l) for d, l in ds]
            random.shuffle(ds)

        for data, label in ds:
            if not isinstance(label, torch.Tensor):
                label = torch.Tensor([label])
            c = label.item()
            class_counts[c] = class_counts.get(c, 0) + 1
            if class_counts[c] <= k:
                train_data.append(torch.unsqueeze(data, 0))
                train_label.append(torch.unsqueeze(label, 0))
            elif not no_test:
                test_data.append(torch.unsqueeze(data, 0))
                test_label.append(torch.unsqueeze(label, 0))
        train_data = torch.cat(train_data)
        train_label = torch.cat(train_label)

        if not no_test:
            test_data = torch.cat(test_data)
            test_label = torch.cat(test_label)

        if as_np:
            train_data = train_data.numpy()
            train_label = train_label.numpy()

            if not no_test:
                test_data = test_data.numpy()
                test_label = test_label.numpy()
                return train_data, train_label, test_data, test_label 
            return train_data, train_label

        return (TensorDataset(train_data, train_label), 
            TensorDataset(test_data, test_label))

    @staticmethod
    def mnist_datasets(linearize=False, train_test=True):
        mnist_transforms=[
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))]
        if linearize:
            mnist_transforms.append(ReshapeTransform((-1,)))
        train_ds = datasets.MNIST('../data', train=True, download=True, 
            transform=transforms.Compose(mnist_transforms));
        test_ds = datasets.MNIST('../data', train=False, 
            transform=transforms.Compose(mnist_transforms));

        if train_test:
            return train_ds, test_ds
        else:
            return torch.utils.data.ConcatDataset((train_ds, test_ds))

    @staticmethod
    def sample_traintest_by_class(train_ds, test_ds, k=None):  
        if k is not None:
            train_ds_part, test_ds_part = DataUtils.sample_by_class(train_ds, k)
            test_ds_part = test_ds
        else:
            train_ds_part, test_ds_part = train_ds, test_ds

        return train_ds_part, test_ds_part

    @staticmethod
    def split_dataset(ds, train_frac=0.6, test_frac=None, validate_frac=0.0):
        if test_frac is None and validate_frac is None:
            raise ValueError("both test_frac and validation_frac should not be None")

        total_len = len(ds)
        train_len = int(total_len * train_frac)

        test_frac = test_frac or (1 - train_frac - validate_frac)
        test_len = int(total_len * test_frac)

        validate_frac = validate_frac or (1 - train_frac - test_frac)
        validate_len = int(total_len * validate_frac)

        train_len = total_len-validate_len-test_len
        
        if validate_len > 0:
            train_ds, test_ds, validate_ds = torch.utils.data.random_split(ds, 
                (train_len, test_len, validate_len))
            return (train_ds, test_ds, validate_ds)
        else:
            train_ds, test_ds = torch.utils.data.random_split(ds, 
                (train_len, test_len))
            return (train_ds, test_ds)

    @staticmethod
    def get_dataloaders(config, train_ds, test_ds):
        kwargs_train = {'pin_memory': True} if config.train_config.use_cuda else {}
        kwargs_test = {'pin_memory': True} if config.test_config.use_cuda else {}

        train_loader = torch.utils.data.DataLoader(train_ds,
            batch_size=config.train_config.batch_size, shuffle=True, **kwargs_train) \
                if train_ds is not None else None

        test_loader = torch.utils.data.DataLoader(test_ds,
            batch_size=config.test_config.batch_size, shuffle=True, **kwargs_test) \
                if test_ds is not None else None

        return train_loader, test_loader




import math

def num2tuple(num):
    return num if isinstance(num, tuple) else (num, num)

def get_missing_l(l, kernal, pad, stride):
    scans = round((l+pad-k)/stride)
    return kernal + scans * stride - l

def conv2d_output_shape(h_w, kernel_size=1, stride=1, pad=0, dilation=1):
    h_w, kernel_size, stride, pad, dilation = num2tuple(h_w), \
        num2tuple(kernel_size), num2tuple(stride), num2tuple(pad), num2tuple(dilation)
    pad = num2tuple(pad[0]), num2tuple(pad[1])
    
    h = math.floor((h_w[0] + sum(pad[0]) - dilation[0]*(kernel_size[0]-1) - 1) / stride[0] + 1)
    w = math.floor((h_w[1] + sum(pad[1]) - dilation[1]*(kernel_size[1]-1) - 1) / stride[1] + 1)
    
    h_missed = h_w[0]+sum(pad[0])-(kernel_size[0]+(h-1)*stride[0])
    w_missed = h_w[1]+sum(pad[1])-(kernel_size[1]+(w-1)*stride[1])

    return (h, w), (h_missed, w_missed)

def convtrans2d_output_shape(h_w, kernel_size=1, stride=1, pad=0, dilation=1, out_pad=0):
    h_w, kernel_size, stride, pad, dilation, out_pad = num2tuple(h_w), \
        num2tuple(kernel_size), num2tuple(stride), num2tuple(pad), num2tuple(dilation), num2tuple(out_pad)
    pad = num2tuple(pad[0]), num2tuple(pad[1])
    
    h = (h_w[0] - 1)*stride[0] - sum(pad[0]) + dilation[0]*(kernel_size[0]-1) + out_pad[0] + 1
    w = (h_w[1] - 1)*stride[1] - sum(pad[1]) + dilation[1]*(kernel_size[1]-1) + out_pad[1] + 1
    
    return h, w

def conv2d_get_padding(h_w_in, h_w_out, kernel_size=1, stride=1, dilation=1):
    h_w_in, h_w_out, kernel_size, stride, dilation = num2tuple(h_w_in), num2tuple(h_w_out), \
        num2tuple(kernel_size), num2tuple(stride), num2tuple(dilation)
    
    p_h = ((h_w_out[0] - 1)*stride[0] - h_w_in[0] + dilation[0]*(kernel_size[0]-1) + 1)
    p_w = ((h_w_out[1] - 1)*stride[1] - h_w_in[1] + dilation[1]*(kernel_size[1]-1) + 1)
    
    return (math.floor(p_h/2), math.ceil(p_h/2)), (math.floor(p_w/2), math.ceil(p_w/2))

def convtrans2d_get_padding(h_w_in, h_w_out, kernel_size=1, stride=1, dilation=1, out_pad=0):
    h_w_in, h_w_out, kernel_size, stride, dilation, out_pad = num2tuple(h_w_in), num2tuple(h_w_out), \
        num2tuple(kernel_size), num2tuple(stride), num2tuple(dilation), num2tuple(out_pad)
        
    p_h = -(h_w_out[0] - 1 - out_pad[0] - dilation[0]*(kernel_size[0]-1) - (h_w_in[0] - 1)*stride[0]) / 2
    p_w = -(h_w_out[1] - 1 - out_pad[1] - dilation[1]*(kernel_size[1]-1) - (h_w_in[1] - 1)*stride[1]) / 2
    
    return (math.floor(p_h/2), math.ceil(p_h/2)), (math.floor(p_w/2), math.ceil(p_w/2))


import torch.optim.lr_scheduler as lr_scheduler
import math 

class GradientRatioScheduler(lr_scheduler._LRScheduler):
    def __init__(self, callbacks, optimizer, last_epoch=-1, smooth=0, decay_factor=1):
        self.lr_factors = [1 for _ in optimizer.param_groups]
        self.smooth = smooth
        self.decay_factor = decay_factor
        self.cached_lrs = None
        callbacks.after_batch.subscribe(self.on_after_batch)
        super(GradientRatioScheduler, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        if self.cached_lrs is None:
            self.cached_lrs = self.get_rat_mul_no_decay_lr() # self.get_rat_blended_lr() #self.get_rat_mul_lr()
        return self.cached_lrs

    def blend_lr(self, lr0, factor, opacity):
        return opacity*lr0 + (1-opacity)*(lr0*factor)
    def get_rat_blended_lr(self):
        epoch = max(self.last_epoch, 0)
        opacity = min(epoch/20.0, 1.0)
        return [self.blend_lr(base_lr * self.decay_factor, self.lr_factors[i], opacity) for i,base_lr in enumerate(self.base_lrs)]

    def get_shake_it_lr(self):
        return [8*(base_lr * self.decay_factor) if self.last_epoch > 1 and self.last_epoch % 15 == 0 else (base_lr * self.decay_factor) \
           for i,base_lr in enumerate(self.base_lrs)]

    def get_rat_mul_lr(self):
        return [(base_lr * self.decay_factor) * self.lr_factors[i] for i,base_lr in enumerate(self.base_lrs)]
    def get_rat_mul_no_decay_lr(self):
        return [base_lr * self.lr_factors[i] for i,base_lr in enumerate(self.base_lrs)]

    def get_rat_exp_disc_lr(self):
        x = max(self.last_epoch, 1e-10) / 8
        return [(base_lr * self.decay_factor) * \
            (self.lr_factors[i] / math.exp(x) + math.exp(-1.0/x)) \
            for i,base_lr in enumerate(self.base_lrs)]

    def get_decay_factor(self):
        return self.decay_factor
    def set_decay_factor(self, decay_factor):
        self.decay_factor = decay_factor
        self.cached_lrs = None

    def on_after_batch(self, *kargs, **kwargs):
        this_g_sum, this_g_count = 0, 0
        last_g = None

        i, m_i = 0, None
        i_start = None
        pgs = list(self.optimizer.param_groups)
        while i < len(pgs):
            pg = pgs[i]
            if m_i is None:
                m_i = pg['m_i']
            #TODO remove dep on m_i?
            if pg['m_i'] != m_i: # new module started
                # save last moddule's avg
                if last_g is None and this_g_count > 0:
                    last_g = this_g_sum / this_g_count
                this_g_sum, this_g_count = 0, 0
                m_i = pg['m_i']
                i_start = i

            # accumulate for current module
            #TODO - worry about requires_grad
            for p in pg['params']:
                if p.requires_grad and p.grad is not None and p.grad.numel() > 0:
                    this_g_sum += p.grad.abs().sum().item()
                    this_g_count += p.grad.numel()
            
            # if this is the last one in module, update LR
            if last_g is not None and i_start is not None and this_g_count >0 and \
                (i == len(pgs)-1 or pgs[i+1]['m_i'] != m_i):

                rat = last_g / (this_g_sum / this_g_count)

                #rat_sig = math.tanh(rat)
                if math.isnan(rat):
                    rat = 1
                rat = max(min(rat, 20.0), 0.1)

                #if rat < 0.1 or rat > 10 or math.isnan(rat):
                #    print(rat, m_i, i)
                for j in range(i, i_start-1, -1):
                    self.lr_factors[j] = \
                        self.smooth*self.lr_factors[j] + (1-self.smooth)*rat
            i += 1
        self.cached_lrs = None

    @staticmethod
    def get_named_members(model, get_members_fn, prefix='', recurse=True):
        r"""Helper method for yielding various names + members of modules."""
        memo = set()
        modules = model.named_modules(prefix=prefix) if recurse else [(prefix, model)]
        for i, (module_prefix, module) in enumerate(modules):
            members = get_members_fn(module)
            for k, v in members:
                if v is None or v in memo:
                    print("WARNING: reused module parameter")
                    continue
                memo.add(v)
                name = module_prefix + ('.' if module_prefix else '') + k
                yield name, v, i

    @staticmethod
    def named_parameters(model, prefix='', recurse=True):
        gen = GradientRatioScheduler.get_named_members(model,
            lambda module: module._parameters.items(),
            prefix=prefix, recurse=recurse)
        for name, p, i in gen:
            yield name, p, i

    @staticmethod
    def get_opt_params(model, lr):
        param_lr=[]
        for name, p, i in GradientRatioScheduler.named_parameters(model):
            param_lr.append({'params': p, 'm_i':i, 'lr': lr})
        param_lr = list(reversed(param_lr)) # we start from last layer
        return param_lr


from .probe import *
import tensorwatch as tw

class TensorWatchProbe(Probe):
    def __init__(self, exp_name, run_name, epoch_config, model, 
            callbacks, metrics, log_config, port):
        super(TensorWatchProbe, self).__init__(exp_name, run_name, epoch_config, model, 
            callbacks, metrics, log_config)

        tw.utils.set_debug_verbosity(log_config.debug_verbosity)

        self.tw = tw.Watcher(port=port)
        self.metrics = metrics
        self.tw.set_globals(model=model, metrics=metrics.stats)

        #TODO manage client server connections better
        #tw.wait_key("Press any key to continue...")

    def on_after_batch(self, train_test, batch_state):
        super(TensorWatchProbe, self).on_after_batch(train_test, batch_state)
        self.tw.observe(event_name='batch', x=self.metrics.stats.epochf, batch=batch_state, tt=train_test)

    def on_after_epoch(self, train_test, loader):
        super(TensorWatchProbe, self).on_after_epoch(train_test, loader)
        self.tw.end_event('batch')
        self.tw.observe(event_name='epoch', tt=train_test)

    #TODO log end epoch




from torch.nn.modules.loss import _Loss

class WeightedMseLoss(_Loss):
    __constants__ = ['reduction']

    def __init__(self, size_average=None, reduce=None, reduction='mean'):
        super(WeightedMseLoss, self).__init__(size_average, reduce, reduction)

    def forward(self, input, target, weight):
        ele_loss = weight * ((input - target) ** 2)
        row_loss = ele_loss.sum(dim=tuple(range(1, len(ele_loss.shape))))
        weight_sum = weight.sum(dim=tuple(range(1, len(weight.shape))))
        batch_item_loss = row_loss / weight_sum

        if self.reduction=='mean':
            return batch_item_loss.mean()
        elif self.reduction=='none':
            return batch_item_loss
        elif self.reduction=='sum':
            return batch_item_loss.sum()
        else:
            raise ValueError("reduction is not supported: {}".format(reduction))





class Probe:
    def __init__(self, exp_name, run_name, epoch_config, model, 
            callbacks, metrics, log_config):
        self.epoch_config = epoch_config
        self.run_name = run_name
        self.log_config = log_config
        self.metrics = metrics
        self.model = model
        
        callbacks.before_start.subscribe(self.on_before_start)
        callbacks.after_epoch.subscribe(self.on_after_epoch)
        callbacks.after_batch.subscribe(self.on_after_batch)

    def on_before_start(self, train_test, epochs, loader):
        print("[{}] CUDA: {}, Batch Size: {}, Data Size: {}, Batches: {}".format(self.run_name, 
            self.epoch_config.use_cuda, self.epoch_config.batch_size, len(loader.dataset), len(loader)))

    def on_after_batch(self, train_test, batch_state):
        # print("[{}] {}".format(self.run_name, self.metrics.get_after_batch_summary()))
        pass

    def on_after_epoch(self, test_train, loader):
        print("[{}] {}".format(self.run_name, self.metrics.get_after_epoch_summary()))



import torch
import argparse


class Config:
    class CommonEpochConfig:
        def __init__(self, batch_size):
            self.batch_size = batch_size
    class TrainEpochConfig(CommonEpochConfig):
        def __init__(self, batch_size):
            super(Config.TrainEpochConfig, self).__init__(batch_size)
            self.lr = 0.01
            self.momentum = 0.5
    class TestEpochConfig(CommonEpochConfig):
        def __init__(self, batch_size):
            super(Config.TestEpochConfig, self).__init__(batch_size)
            self.batch_size = batch_size
    class LogConfig:
        def __init__(self, model_graph=False, false_preds=False, param_histo_freq=0,
                     param_histo_next_epoch=0, basic=False, debug_verbosity=-1):
            self.model_graph = model_graph
            self.false_preds = false_preds
            self.param_histo_freq = param_histo_freq
            self.param_histo_next_epoch = param_histo_next_epoch
            self.basic = basic
            self.debug_verbosity = debug_verbosity

    def __init__(self, args):
        self.args = args
        self.train_config = Config.TrainEpochConfig(args.batch_size)
        self.test_config = Config.TestEpochConfig(args.test_batch_size)
        self.log_config = Config.LogConfig(debug_verbosity=args.debug_verbosity)
        self.epochs = args.epochs
        self.seed = args.seed


    @staticmethod
    def from_args(train_batch_size=64, test_batch_size=1000, 
                  learning_rate=0.01, momentum=0.0, weight_decay=0, debug_verbosity=None,
                  rand_seed=42,no_cuda_train=False, no_cuda_test=False, epochs=100,
                  arg_parser:argparse.ArgumentParser=None):

        parser = arg_parser or argparse.ArgumentParser(description='PyTorch Deep Learning Pipeline')
        parser.add_argument('--batch-size', type=int, default=train_batch_size, #64
                            help='input batch size for training')
        parser.add_argument('--test-batch-size', type=int, default=test_batch_size,
                            help='input batch size for testing')
        parser.add_argument('--epochs', type=int, default=epochs,
                            help='number of epochs to train')
        parser.add_argument('--lr', type=float, default=learning_rate, #0.01
                            help='learning rate')
        parser.add_argument('--momentum', type=float, default=momentum, 
                            help='SGD momentum')
        parser.add_argument('--weight-decay', type=float, default=weight_decay,
                            help='Weight Decay')
        parser.add_argument('--no-cuda-train', action='store_true', default=no_cuda_train,
                            help='disables CUDA training')
        parser.add_argument('--no-cuda-test', action='store_true', default=no_cuda_test,
                            help='disables CUDA testing')
        parser.add_argument('--seed', type=int, default=rand_seed,
                            help='random seed (default: 42)')
        parser.add_argument('--debug-verbosity', type=int, default=debug_verbosity,
                            help='Debug Verbosity (default: -1)')


        args = parser.parse_args()

        config = Config(args)

        config.train_config.use_cuda = not args.no_cuda_train and torch.cuda.is_available()
        config.train_config.device = torch.device("cuda" if config.train_config.use_cuda else "cpu")
        config.train_config.lr = args.lr
        config.train_config.momentum = args.momentum
        config.train_config.weight_decay = args.weight_decay
        
        config.test_config.use_cuda = not args.no_cuda_test and torch.cuda.is_available()
        config.test_config.device = torch.device("cuda" if config.test_config.use_cuda else "cpu")

        if config.seed is not None:
            torch.manual_seed(config.seed)

        return config



import timeit

class Metrics:
    class Stats:
        def __init__(self):
            self.reset()
        def reset(self):
            self.epoch_index = -1
            self.epoch_loss:float = None
            self.epoch_time = 0.
            self.batch_index = -1
            self.batch_loss:float = None
            self.batch_time = 0.
            self.loss_sum = 0.
            self.correct_sum = 0
            self.epoch_count = 0
            self.batch_count = 0
            self.epochf = 0

    def __init__(self, callbacks):
        self.stats = Metrics.Stats()
        callbacks.before_start.subscribe(self.on_before_start)
        callbacks.before_batch.subscribe(self.on_before_batch)
        callbacks.before_epoch.subscribe(self.on_before_epoch)
        callbacks.after_batch.subscribe(self.on_after_batch)
        callbacks.after_epoch.subscribe(self.on_after_epoch)

    def on_before_start(self, train_test, epochs, loader):
        self.stats.reset()
        self.stats.epoch_count = epochs

    def on_before_epoch(self, train_test, loader):
        self.epoch_start_time = timeit.default_timer()
        self.stats.loss_sum = 0
        self.stats.batch_index = -1
        self.stats.epoch_index += 1
        self.stats.epochf = self.stats.epoch_index
        self.stats.batch_count = len(loader)

    def on_before_batch(self, train_test, batch_state):
        self.batch_start_time = timeit.default_timer()
        self.stats.batch_index += 1
        self.stats.epochf = self.stats.epoch_index + \
            (float(self.stats.batch_index) / self.stats.batch_count)

    def on_after_batch(self, test_train, batch_state):
        self.stats.batch_time = timeit.default_timer() - self.epoch_start_time
        self.stats.loss_sum += batch_state.loss_all.sum().item()
        self.stats.batch_loss = batch_state.loss
        self.stats.batch_loss_all = batch_state.loss_all

    def on_after_epoch(self, test_train, loader):
        self.stats.epoch_time = timeit.default_timer() - self.epoch_start_time
        self.stats.epoch_loss = self.stats.loss_sum / len(loader.dataset)

    def get_after_batch_summary(self):
        return "Batch: {}, loss: {:.4f}, Time: {:.2f}".format( \
            self.stats.epochf, self.stats.batch_loss, self.stats.batch_time)
    def get_after_epoch_summary(self):
        return "Epoch: {}, loss: {:.4f}, Time: {:.2f}".format( \
            self.stats.epoch_index, self.stats.epoch_loss, self.stats.epoch_time)

class ClassificationMetrics(Metrics):
    def __init__(self, *args,**kwargs):
        super(ClassificationMetrics, self).__init__(*args,**kwargs)
        self.stats.correct_sum = 0
        self.stats.epoch_accuracy = 0.
        self.stats.batch_accuracy = 0.

    def on_before_epoch(self, train_test, loader):
        super(ClassificationMetrics, self).on_before_epoch(train_test, loader)
        self.stats.correct_sum = 0

    def on_after_batch(self, test_train, batch_state):
        super(ClassificationMetrics, self).on_after_batch(test_train, batch_state)
        # get the index of the max log-probability
        pred = batch_state.output.max(1, keepdim=True)[1]
        batch_correct = pred.eq(batch_state.target.view_as(pred)).sum().item()
        self.stats.correct_sum += batch_correct
        self.stats.batch_accuracy = float(batch_correct) / len(batch_state.input)

    def on_after_epoch(self, test_train, loader):
        super(ClassificationMetrics, self).on_after_epoch(test_train, loader)
        self.stats.epoch_accuracy = self.stats.correct_sum / len(loader.dataset)

    def get_after_batch_summary(self):
        s = super(ClassificationMetrics, self).get_after_batch_summary()
        return "{}, accuracy: {:.4f}".format( \
            s, self.stats.batch_accuracy)
    def get_after_epoch_summary(self):
        s = super(ClassificationMetrics, self).get_after_epoch_summary()
        return "{}, accuracy: {:.4f}".format( \
            s, self.stats.epoch_accuracy)


        

